//VGM Player with SN76489 sound chip emulation v1.12
//by Shiru (shiru@mail.ru) 01'10

package vgmplayer
{
import flash.utils.*;
import flash.media.*;
import flash.events.*;

public class VGMPlayer
{
    private var data:ByteArray;
    private var sound:Sound;
    private var channel:SoundChannel;
    private var chip:SN76489;
    private var renderActive:Boolean;
    private var wait:uint;
    private var ptr:uint;
    private var loop:uint;
    private var enable:Boolean;
    private var looping:Boolean;
    private var trackName:String;
    private var gameName:String;
    private var systemName:String;
    private var originalAuthor:String;
    private var releaseDate:String;
    private var whoConverted:String;

 
    public function VGMPlayer():void
    {
        chip=new SN76489();
        unload();
        setLooping(true);

        sound=new Sound();
        sound.addEventListener(SampleDataEvent.SAMPLE_DATA,soundUpdate);
        channel=sound.play();
    }
 
    private function setAllInfo(str:String):void
    {
        trackName=str;
        gameName=str;
        systemName=str;
        originalAuthor=str;
        releaseDate=str;
        whoConverted=str;
    }
 
    private function resetInfo():void
    {
        setAllInfo("No file loaded");
    }

    private function readInt(off:uint):int
    {
        return (data[off]+(data[off+1]<<8)+(data[off+2]<<16)+(data[off+3]<<24));
    }
 
    private function readString(off:uint):String
    {
        var str:String;
        var char:uint;

        str="";

        while(true)
        {
            char=data[off]+(data[off+1]<<8);
            off+=2;
            if(!char) break;
            str+=String.fromCharCode(char);
        }
 
        return str;
    }
 
    private function skipString(off:uint):uint
    {
        var char:uint;
 
        while(true)
        {
            char=data[off]+(data[off+1]<<8);
            off+=2;
            if(!char) break;
        }
 
        return off;
    }
 
    public function load(v:Class):Boolean
    {
        const unknown:String="Unknown";
        var i:uint,off:uint,len:uint;
        var char:uint;
        var temp:ByteArray;

        unload();

        if(!v) return false;
        data=new v();
        if(!data) return false;
 
        if(data[0]==0x1f&&data[1]==0x8b&&data[2]==0x08)
        {
            off=10;
            len=data.length;
            char=data[3];
            if(char&0x04) off+=(data[13]+(data[12]<<8)+4);
            if(char&0x08)
            {
                while(data[off]) off+=1;
                off+=1;
            }
            if(char&0x10)
            {
                while(data[off]) off+=1;
                off+=1;
            }
            if(char&0x02) off+=2;
 
            temp=new ByteArray();
            while(off<len)
            {
            	temp.writeByte(data[off]);
            	off++;
            }
            data=temp;
            temp=null;
            data.inflate();
        }
 
        if(readInt(0)!=0x206d6756) return false;
 
        setAllInfo("No GD3 tag found");
        loop=readInt(0x1c)+0x1c;
        chip.clock(readInt(0x0c));
        off=readInt(0x14);

        if(off)
        {
            if(readInt(off+0x14)!=0x20336447) return true;
            if(readInt(off+0x18)!=0x00000100) return true;
 
            if(off)
            {
                off+=0x20;

                trackName=readString(off);
                off=skipString(off);
                off=skipString(off);
                gameName=readString(off);
                off=skipString(off);
                off=skipString(off);
                systemName=readString(off);
                off=skipString(off);
                off=skipString(off);
                originalAuthor=readString(off);
                off=skipString(off);
                off=skipString(off);
                releaseDate=readString(off);
                off=skipString(off);
                whoConverted=readString(off);
                off=skipString(off);
 
                if(trackName=="")      trackName=unknown;
                if(gameName=="")       gameName=unknown;
                if(systemName=="")     systemName=unknown;
                if(originalAuthor=="") originalAuthor=unknown;
                if(releaseDate=="")    releaseDate=unknown;
                if(whoConverted=="")   whoConverted=unknown;
            }
        }

        return true;
    }
 
    public function unload():void
    {
        stop();
        data=null;
        resetInfo();
    }

    public function play():void
    {
        chip.reset();
        ptr=0x40;
        enable=true;
    }
 
    public function stop():void
    {
        enable=false;
        while(renderActive);
        chip.reset();
    }
 
    public function getTrackName():String
    {
        return trackName;
    }

    public function getGameName():String
    {
        return gameName;
    }

    public function getSystemName():String
    {
        return systemName;
    }

    public function getOriginalAuthor():String
    {
        return originalAuthor;
    }

    public function getReleaseDate():String
    {
        return releaseDate;
    }

    public function getWhoConverted():String
    {
        return whoConverted;
    }

    public function setLooping(l:Boolean):void
    {
        looping=l;
    }

    public function setSoundTransform(s:SoundTransform):void
    {
        channel.soundTransform=s;
    }
 
    public function render(buf:ByteArray,len:uint):void
    {
        var i:uint;
        var tag:uint;
        var inc:uint;
 
        if(!enable)
        {
            for(i=0;i<len;i+=1)
            {
                buf.writeFloat(0);
                buf.writeFloat(0);
            }
        }
        else
        {
            i=0;
 
            if(wait)
            {
                if(wait>=len)
                {
	                chip.render(buf,len);
	                wait-=len;
	                return;
                }
                else
                {
	                chip.render(buf,wait);
	                i=wait;
                }
            }

            while(i<len)
            {
                wait=0;
 
                while(!wait)
                {
                    tag=data[ptr];
                    switch(tag)
                    {
                    case 0x4f:
                        inc=2;
                        break;
                    case 0x50:
                        chip.write(data[ptr+1]);
                        inc=2;
                        break;
                    case 0x61:
                        wait=data[ptr+1]+(data[ptr+2]<<8);
                    case 0x51:
                    case 0x52:
                    case 0x53:
                    case 0x54:
                        inc=3;
                        break;
                    case 0x62:
                        wait=735;
                        inc=1;
                        break;
                    case 0x63:
                        wait=882;
                        inc=1;
                        break;
                    case 0x66:
                        if(looping)
                        {
                            if(readInt(0x20)) ptr=loop; else ptr=0x40;
                        }
                        else
                        {
                            enable=false;
                            chip.reset();
                            wait=10000;
                        }
                        inc=0;
                        break;
                    case 0x67:
                        inc=1;
                        break;
                    default:
                        if(tag>=0x70&&tag<0x80)
                        {
                            wait=(tag&0x0f)+1;
                            inc=1;
                            break;
                        }
                        if(tag>=0x30&&tag<0x4f)
                        {
                            inc=2;
                            break;
                        }
                        if((tag>=0x55&&tag<0x60)||(tag>=0xa0&&tag<0xc0))
                        {
                            inc=3;
                            break;
                        }
                        if(tag>=0xc0&&tag<0xe0)
                        {
                            inc=4;
                            break;
                        }
                        if(tag>=0xe1)
                        {
                            inc=5;
                            break;
                        }
                        inc=2;
                    }
                    ptr+=inc;
                }

                if(i+wait>len)
                {
                    chip.render(buf,len-i);
                    wait-=(len-i);
                    break;
                }
                chip.render(buf,wait);
                i+=wait;
            }
        }
    }

    private function soundUpdate(e:SampleDataEvent):void
    {
        renderActive=true;
        render(e.data,8192);
        renderActive=false;
    }
}
}